@@ -330,7 +330,7 @@ soup_cookie_jar_get_cookies (SoupCookieJar *jar, SoupURI *uri,
 	priv = SOUP_COOKIE_JAR_GET_PRIVATE (jar);
 	g_return_val_if_fail (uri != NULL, NULL);
 
-	if (!SOUP_URI_VALID_FOR_HTTP (uri))
+	if (!uri->host)
 		return NULL;
 
 	/* The logic here is a little weird, but the plan is that if
@@ -486,7 +486,7 @@ soup_cookie_jar_set_cookie (SoupCookieJar *jar, SoupURI *uri,
 	g_return_if_fail (uri != NULL);
 	g_return_if_fail (cookie != NULL);
 
-	if (!SOUP_URI_VALID_FOR_HTTP (uri))
+	if (!uri->host)
 		return;
 
 	priv = SOUP_COOKIE_JAR_GET_PRIVATE (jar);
@@ -530,7 +530,7 @@ soup_cookie_jar_set_cookie_with_first_party (SoupCookieJar *jar,
 	g_return_if_fail (first_party != NULL);
 	g_return_if_fail (cookie != NULL);
 
-	if (!SOUP_URI_VALID_FOR_HTTP (uri))
+	if (!uri->host)
 		return;
 
 	priv = SOUP_COOKIE_JAR_GET_PRIVATE (jar);
@@ -563,11 +563,8 @@ process_set_cookie_header (SoupMessage *msg, gpointer user_data)
 	for (nc = new_cookies; nc; nc = nc->next) {
 		SoupURI *first_party = soup_message_get_first_party (msg);
 		
-		if (first_party == NULL &&
-		    priv->accept_policy == SOUP_COOKIE_JAR_ACCEPT_NO_THIRD_PARTY)
-			continue; /* Can't check anything */
-
 		if ((priv->accept_policy == SOUP_COOKIE_JAR_ACCEPT_NO_THIRD_PARTY &&
+		     first_party != NULL && first_party->host &&
 		     soup_cookie_domain_matches (nc->data, first_party->host)) ||
 		    priv->accept_policy == SOUP_COOKIE_JAR_ACCEPT_ALWAYS)
 			soup_cookie_jar_add_cookie (jar, nc->data);
