@@ -445,6 +445,24 @@ test_disabled (const char *path)
 	g_main_loop_unref (loop);
 }
 
+/* Fix up XDG_DATA_DIRS for jhbuild runs so that it still works even
+ * if you didn't install shared-mime-info.
+ */
+static void
+fixup_xdg_dirs (void)
+{
+	const char *xdg_data_dirs = g_getenv ("XDG_DATA_DIRS");
+	char *new_data_dirs;
+
+	if (xdg_data_dirs &&
+	    !g_str_has_prefix (xdg_data_dirs, "/usr/share") &&
+	    !strstr (xdg_data_dirs, ":/usr/share")) {
+		new_data_dirs = g_strdup_printf ("%s:/usr/share", xdg_data_dirs);
+		g_setenv ("XDG_DATA_DIRS", new_data_dirs, TRUE);
+		g_free (new_data_dirs);
+	}
+}
+
 int
 main (int argc, char **argv)
 {
@@ -452,6 +470,8 @@ main (int argc, char **argv)
 
 	test_init (argc, argv, NULL);
 
+	fixup_xdg_dirs ();
+
 	server = soup_test_server_new (TRUE);
 	soup_server_add_handler (server, NULL, server_callback, NULL, NULL);
 	base_uri = soup_uri_new ("http://127.0.0.1/");
@@ -540,7 +560,7 @@ main (int argc, char **argv)
 	test_sniffing ("/multiple_headers/home.gif", "image/gif");
 
 	/* Test that we keep the parameters when sniffing */
-	test_sniffing ("/type/text_html; charset=\"UTF-8\"/test.html", "text/html; charset=\"UTF-8\"");
+	test_sniffing ("/type/text_html; charset=UTF-8/test.html", "text/html; charset=UTF-8");
 
 	/* Test that disabling the sniffer works correctly */
 
@@ -549,6 +569,7 @@ main (int argc, char **argv)
 	soup_uri_free (base_uri);
 
 	soup_test_session_abort_unref (session);
+	soup_test_server_quit_unref (server);
 	test_cleanup ();
 	return errors != 0;
 }
